/* ---------------------------------------------------------- 
%   (C)1993,1994 Institute for New Generation Computer Technology 
%       (Read COPYRIGHT for detailed information.) 
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

/*
 * predicate
 *
 * method table
 * ref to module
 * predicate desc (atom)
 */

#include <stdio.h>
#include <klic/gdobject.h>
#include <klic/predinfo.h>
#include "atom.h"
#include "funct.h"  /* arities */
#include <klic/gmodule.h>
#include <klic/generic.h>  /* isclass */

#define CLASS_NAME  predicate
#define G_OBJ_TYPE  struct predicate_object
#define G_OBJ_SIZE  G_SIZE_IN_Q(G_OBJ_TYPE)

#include <klic/gd_macro.h>

extern const struct modinfo defined_modules[];

GD_USE_CLASS(vector);
GD_USE_CLASS(module);

GDDEF_GUNIFY()
{
  return (g_self->method_table == GD_OTHER->method_table &&
	  g_self->pdesc == GD_OTHER->pdesc ) ?
    GENERIC_SUCCEEDED : GENERIC_FAILED;
}

GDDEF_UNIFY()
{
  G_STD_DECL;

  if( g_self->method_table != GD_OTHER->method_table ||
      g_self->pdesc != GD_OTHER->pdesc )
    GD_UNIFY_FAIL;
}

G_DEF_GC()
{
  G_STD_DECL;
  G_OBJ_TYPE* newself;

  G_SET_NEWOBJ_IN_NEWGEN(newself);
  newself->pdesc = g_self->pdesc;
  G_COPY_KL1_TERM_TO_NEWGEN(g_self->module_obj, newself->module_obj);
  newself->predicate_name = g_self->predicate_name;
  G_RETURN_FROM_GC(newself);
}

/* Generic method */

GDDEF_METHOD(module_1)
{
  G_STD_DECL;

  do_shallow_unify(g_argv[0], g_self->module_obj);
}

GDDEF_METHOD(name_1)
{
  G_STD_DECL;

  do_shallow_unify(g_argv[0], g_self->predicate_name);
}

GDDEF_METHOD(arity_1)
{
  G_STD_DECL;
  const struct predicate* pred = g_self->pdesc;

  do_shallow_unify(g_argv[0], makeint(pred->arity));
}

#define GD_ALLOC_GOAL(goal, pdesc, size) \
do{ \
  (goal) = (struct goalrec*) klic_alloc((size) + 2); \
  (goal)->pred = (pdesc); \
}while(0)

#define GD_GOAL_ARG(i, goal)  ((goal)->args[i])

GDDEF_METHOD(apply_1)
{
  G_STD_DECL;
  struct goalrec* goal;
  q argv[2];
  int size;
  int i;
  q func;

  GD_DEREF(g_argv[0]);
  func = g_argv[0];
  if( !isfunctor(func) || !isclass(func, vector) ){
    GD_ERROR_IN_METHOD("Invalid argument specification", "apply");
  }

  GD_CALL_GMETHOD(func, vector_1, argv);
  size = intval(argv[0]);
  if( g_self->pdesc->arity != size ){
    GD_ERROR_IN_METHOD("Arity mismatch", "apply");
  }
  GD_ALLOC_GOAL(goal, g_self->pdesc, size);
  for( i=0; i<size; ++i ){
    argv[0] = makeint(i);
    GD_CALL_GMETHOD(func, element_2, argv);
    GD_GOAL_ARG(i, goal) = argv[1];
  }
  resume_same_prio(goal);
}

#define GD_METHOD_ARITY  arities (g_method_functor)

GDDEF_METHOD(call)
{
  G_STD_DECL;
  struct goalrec* goal;
  int i;
  int size = GD_METHOD_ARITY;

  GD_ALLOC_GOAL(goal, g_self->pdesc, size);
  for( i=0; i<size; ++i ){
    GD_GOAL_ARG(i, goal) = g_argv[i];
  }
  resume_same_prio(goal);
}

GDDEF_GENERIC()
{
  G_STD_DECL;

  switch( g_method_functor ){
    GD_METHOD_CASE(module_1);
    GD_METHOD_CASE(name_1);
    GD_METHOD_CASE(arity_1);
    GD_METHOD_CASE(apply_1);
  default:
    switch( functors(g_method_functor) ){
      GD_METHOD_NAME_CASE(call);
      GD_METHOD_CASE_DEFAULT;
    }
  }
}

G_DEF_PRINT()
{
  G_STD_DECL;
  G_PRINT("predicate#(");
  GD_PRINT_KL1_TERMS
    (((struct module_object*) functorp(g_self->module_obj))->name, 0, 1);
  G_PRINT(":");
  GD_PRINT_KL1_TERMS(g_self->predicate_name, 0, 1);
  GD_PRINTF("/%d)", g_self->pdesc->arity);
  G_RETURN_FROM_PRINT();
}

GDDEF_GMETHOD(arity_1)
{
  G_STD_DECL;
  const struct predicate* pred = g_self->pdesc;

  g_argv[0] =  makeint(pred->arity);
  return GENERIC_SUCCEEDED;
}

GDDEF_GMETHOD(predicate_0)
{
  return (g_self->pdesc ? GENERIC_SUCCEEDED : GENERIC_FAILED);
}

GDDEF_GGENERIC()
{
  switch( g_method_functor ){
    GD_GMETHOD_CASE(predicate_0);
    GD_GMETHOD_CASE(arity_1);
    GD_GMETHOD_CASE_DEFAULT;
  }
}

#define GDUSE_MY_GUNIFY
#define GDUSE_MY_UNIFY
#define GDUSE_MY_PRINT
#define GDUSE_MY_GC
#define GDUSE_MY_GENERIC
#define GDUSE_MY_GGENERIC

#include <klic/gd_methtab.h>

extern const struct predicate*
locate_predicate_in_module(mname, pname, arity)
  q mname, pname;
  long arity;
{
  unsigned char* modname = namestringof(mname);
  unsigned char* predname = namestringof(pname);
  const struct modinfo* mp;
  for( mp = defined_modules; mp->name != NULL; mp++ ){
    if( strcmp((char*) mp->name, (char*) modname) == 0 ){
      const struct predinfo* const* pp;
      for( pp = mp->preds; *pp != NULL; pp++ ){
	if( strcmp((char*) ((*pp)->name), (char*) predname) == 0 &&
	    (*pp)->predstruct->arity == arity ){
	  return (*pp)->predstruct;
	}
      }
      return NULL;
    }
  }
  return NULL;
}

/*
  new predicate
*/

G_DEF_NEW()
{
  G_STD_DECL_FOR_NEW();
  G_OBJ_TYPE* newpred;
  struct module_object* module_obj;
  const struct predicate* predaddr;
  q predname;
  int arity;

  if( g_argc != 3 ) G_ERROR_IN_NEW("Too few or too many arguments");

  G_DEREF_FOR_NEW(g_argv[0]);
  if( !isfunctor(g_argv[0]) ||
      (((struct module_object*) functorp(g_argv[0]))->method_table
      != &G_method_table0(module) ) ){
    G_ERROR_IN_NEW("First parameter is not a module");
  }
  module_obj = (struct module_object*) functorp(g_argv[0]);

  GDSET_SYMARG_FOR_NEW(predname, g_argv[1]);
  GDSET_INTARG_FOR_NEW(arity, g_argv[2]);

  predaddr = locate_predicate_in_module(module_obj->name, predname, arity);
  if( predaddr == 0 ){
    char message[1024];
    sprintf(message,
	    "Nonexistent predicate %s:%s/%d",
	    namestringof(module_obj->name),
	    namestringof(predname),
	    arity);
    G_ERROR_IN_NEW(message);
  }

  G_SET_NEWOBJ_FOR_NEW(newpred, sizeof(struct predicate_object));
  newpred->pdesc = predaddr;
  newpred->module_obj = makefunctor(module_obj);
  newpred->predicate_name = predname;
  return makefunctor(newpred);
}
