/* ---------------------------------------------------------- 
%   (C)1993,1994,1995 Institute for New Generation Computer Technology 
%       (Read COPYRIGHT for detailed information.) 
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

/* G_error, fatal, fatalf, debug_printf, klic_fprintf */
#include <klic/gdobject.h>

#include <klic/functorstuffs.h>  /* arities */
#include <klic/atomstuffs.h>
#include <stdio.h>
#include <klic/gd_macro.h>
#include <klic/susp.h>
#include <klic/distpkt.h>  /* combuf */

#define G_NEW_GOAL &predicate_generic_xnew_3
#define GD_GENERIC_GOAL &predicate_generic_xgeneric_2

extern void suspend_goal();
extern q gd_new_pointer();
extern q termarray_g_new(long g_argc, q* g_argv);

extern const struct predicate predicate_generic_xgeneric_2;

/*************************** inlines *******************************/

extern void
G_SUSPEND(x, goal)
  q x;
  struct goalrec* goal;
{
  G_STD_DECL;
  suspend_goal(goal, x);
}

extern void
GD_MAKE_GENERIC_GOAL(goalp1, var, method_functor, argv)
  struct goalrec** goalp1;
  q var;
  unsigned long method_functor;
  q argv[];
{
  G_STD_DECL;
  unsigned long argc = arities (method_functor);
  q objp;
  int i;

  objp = makefunctor(klic_alloc(G_SIZE_IN_Q(struct functor) + argc - 1));
  functor_of(objp) = makesym(method_functor);
  for( i=0; i<argc; i++ ){
    arg(objp, i) = argv[i];
  }

  *goalp1 = (struct goalrec*) klic_alloc(G_SIZE_IN_Q(struct goalrec) - 4);
  (*goalp1)->pred    = GD_GENERIC_GOAL;
  (*goalp1)->args[0] = objp;
  (*goalp1)->args[1] = var;
}

extern void
G_MAKE_THE_NEW_GOAL(var, goalp1, myself, argc, argv)
  q* var;
  struct goalrec** goalp1;
  q (*myself)();
  unsigned long argc;
  q argv[];
{
  G_STD_DECL;
  q newvar;
  q newobj, newobj2;

  newobj = gd_new_pointer(myself);
  newobj2 = termarray_g_new(argc, argv);

  *goalp1 = (struct goalrec*) klic_alloc(G_SIZE_IN_Q(struct goalrec) - 3);
  G_MAKE_VAR(newvar);
  (*goalp1)->pred    = G_NEW_GOAL;
  (*goalp1)->args[0] = newvar;
  (*goalp1)->args[1] = newobj;
  (*goalp1)->args[2] = newobj2;
  *var = newvar;
}

/**********************************************************************/
#define GD_DEREF_FOR_GENERIC(obj) \
do{ \
  while( isref(obj) ){ \
    q temp0 = derefone(obj); \
    if( isref(temp0) && (obj) == derefone(temp0) ){ \
      struct goalrec* goal; \
      GD_MAKE_GENERIC_GOAL(&goal, (obj), g_method_functor, g_argv); \
      G_SUSPEND((obj), goal); \
      return; \
    }else{ \
      (obj) = temp0; \
    } \
  } \
}while(0)



/**********************************************************************/
extern void
gd_generic(g_object, g_method_functor, g_argv)
  q g_object;
  long g_method_functor;
  q* g_argv;
{
  G_STD_DECL;
  struct data_object* g_obj;

  GD_DEREF_FOR_GENERIC(g_object);
  if( !isfunctor(g_object) || !isref(functor_of(g_object)) ){
    klic_printf("### method %F issued to %k ###\n",
		g_method_functor, g_object);
    fatal("Generic method issued to a non-object");
  }
  g_obj = data_objectp(g_object);
  g_obj->method_table->body_generic(g_obj, g_method_functor, g_argv);
}


/**********************************************************************/
extern q
GC_MAKE_HOOK_VAR(obj)
  struct consumer_object* obj;
{
  declare_globals;
  q newvar;
  struct susprec* susp;

  allocnewsusp(newvar, susp);
  susp->u.first_hook.u.o = tag_consumer_hook(obj);
  return newvar;
}

/************************* standard methods *************************/

extern q
GD_STD_GUNIFY(struct data_object* g_self, struct data_object* GD_OTHER)
{
  return GENERIC_SUCCEEDED;
}

extern q*
GD_STD_UNIFY(struct data_object* g_self, struct data_object* GD_OTHER)
{
  return heapp();
}

extern long G_STD_PRINT(
  struct object* g_self, FILE* g_fp,
  unsigned long g_depth, unsigned long g_length )
{
  klic_fprintf(g_fp, "<OBJ@%x>", (long) g_self);
  return 1L;
}

extern q* G_STD_GC(struct object* g_self)
{
  return heapp();
}

extern void G_STD_DEALLOCATE(struct object* g_self)
{
}


extern q*
GD_STD_GENERIC(struct data_object* g_self, long g_method_functor, q* g_argv)
{
  return heapp();
}

extern q
GD_STD_GGENERIC(struct data_object* g_self, long g_method_functor, q* g_argv)
{
  return GENERIC_SUCCEEDED;
}

extern q
GD_STD_COMPARE(struct data_object* g_self, struct data_object* GD_OTHER)
{
  debug_printf("### %k compared with %k ###\n",
	       makefunctor(g_self), makefunctor(GD_OTHER));
  fatal("Comparision of objects of an inappropriate class made");
}

extern q
GD_STD_HASH(struct data_object* g_self, long level)
{
  return makeint((long) g_self->method_table);
}

extern NeverReturn void
G_error(errmsg, where, object_kind, class)
     char *errmsg, *where, *object_kind, *class;
{
  fatalf("%s\n\t(in %s of %s object of class \"%s\")",
	 errmsg, (where != NULL ? where : "unknown method"),
	 object_kind, (class != NULL ? class : ""));
}

extern q G_STD_ENCODE(struct object* self, void* buffer)
{
  return GENERIC_FAILED;
}

/*
  Generator object
*/

extern q
GG_MAKE_HOOK_VAR(obj)
  struct generator_object* obj;
{
  declare_globals;
  q newvar;
  struct generator_susp* gsusp;

  allocnewgensusp(newvar, gsusp);
  gsusp->u.o = tag_generator_susp(obj);
  return newvar;
}

extern q
GG_STD_SUSPEND()
{
  return makecons(0);
}
