/* ----------------------------------------------------------
%   (C)1993,1994 Institute for New Generation Computer Technology
%       (Read COPYRIGHT for detailed information.)
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

#include <stdio.h>
#include <klic/alloc.h>  /* malloc_check */

/* debug_printf, klic_getc, klic_fread, klic_putc, klic_fwrite, klic_fflush */
#include <klic/gcobject.h>

#include "g_pointer.h"
#include <klic/g_string.h>
#include <klic/gd_macro.h>
#ifdef DIST
#include <klic/distio.h>
#endif
#include "atom.h"
#include "funct.h"

#define CLASS_NAME  file__io
#define G_OBJ_TYPE  struct file_io_object
#define G_OBJ_SIZE  G_SIZE_IN_Q(G_OBJ_TYPE)

#include <klic/gc_macro.h>

#define CheckInput() \
do{ \
  if( g_self->infile == NULL ) goto message_error; \
  if( g_self->outfile != NULL ) do_flush(g_self); \
}while(0)

#define CheckOutput() \
do{ \
  if( g_self->outfile == NULL ) goto message_error; \
}while(0)

extern size_t strlen();
extern unsigned char* generic_string_body();
extern q convert_binary_c_string_to_klic_string();

struct file_io_object {
  struct consumer_object_method_table* method_table;
  int linecount;
  q inname, outname;
  q stream; /* saved input stream for suspension by some other reasons */
  FILE* infile;
  FILE* outfile;
};

GD_USE_CLASS(byte__string);
GD_USE_CLASS(pointer);

/* basic method definitions */

static long
do_flush(obj)
  struct file_io_object* obj;
{
#ifdef DIST
  if( obj->infile == stdin ||
      obj->outfile == stdout || obj->outfile == stderr ){
    return UserFflush();
  }else
#endif
  {
    return klic_fflush(obj->outfile);
  }
}

GCDEF_UNIFY()
{
  G_STD_DECL;
  q newvar;
  q reason;

  if( g_self->stream != 0 ){
    g_term = g_self->stream;
    g_self->stream = 0;
  }
 top:
  if( iscons(g_term) ){
    q message;
    GCSET_MESSAGE(message);
    if( isint(message) ){
      long c = intval(message);
#ifdef DIST
      if( g_self->outfile == stdout ){
	UserPutc(c);
      }else if( g_self->outfile == stderr ){
	UserEPutc(c);
      }else
#endif
      if( klic_putc(c, g_self->outfile) == EOF ){
	fatal("putc failed");
      }
    }else if( isfunctor(message) ){
      switch( symval(functor_of(message)) ){

	/*** Input Messages ***/

      case functor_getc_1: {
	long c;
	CheckInput();
#ifdef DIST
	if( g_self->infile == stdin ){
          c = UserGetc();
	}else
#endif
	{
	  c = klic_getc(g_self->infile);
	}
	if( c=='\n' ){ g_self->linecount++; }
	G_UNIFY_VALUE(arg(message, 0), makeint(c));
	break;
      }

      case functor_ungetc_1: {
	long c;
	CheckInput();
	GCSET_MESSAGE_INT_ARG(c, message, 0);
	if( c=='\n' ){ g_self->linecount--; }
#ifdef DIST
	if( g_self->infile == stdin ){
	  UserUngetc(c);
	}else
#endif
	{
	  ungetc(c, g_self->infile);
	}
	break;
      }

      case functor_fread_2: {
	long n, k;
	char* buf;
	q string;
	long space_needed;
	CheckInput();
	GCSET_MESSAGE_INT_ARG(n, message, 0);
	if( n<0 ) goto message_error;
	space_needed = sizeof(struct byte_string_object) + n + sizeof(q);
	if( (char*) heapp()+space_needed >= (char*) real_heaplimit() ){
	  add_this_more_space((space_needed + sizeof(q) - 1) / sizeof(q));
	  goto gc_request;
	}
	buf = (char*) malloc_check(n);
#ifdef DIST
	if( g_self->infile == stdin ){
          Read(buf);
	  n = strlen(buf);
	}else
#endif
	{
	  n = klic_fread(buf, 1, n, g_self->infile);
	}

	for( k=0; k<n; k++ ){
	  if( buf[k] == '\n' ){ g_self->linecount++; }
	}
	string = convert_binary_c_string_to_klic_string(buf, n);
	if( isref(string) ){
	  fatal("internal error: string allocation for fread");
	}
	free(buf);
	G_UNIFY_VALUE(arg(message, 1), string);
	break;
      }

      case functor_linecount_1: {
	CheckInput();
	G_UNIFY_VALUE(arg(message, 0), makeint(g_self->linecount));
	break;
      }

	/*** Output Messages ***/

      case functor_putc_1: {
	long c;
	CheckOutput();
	GCSET_MESSAGE_INT_ARG(c, message, 0);
#ifdef DIST
	if( g_self->outfile == stdout ){ UserPutc((int) c); }
	else if( g_self->outfile == stderr ){ UserEPutc((int) c); }
	else
#endif
	if( klic_putc(c, g_self->outfile) == EOF ){
	  fatal("putc failed");
	}
	break;
      }

      case functor_fwrite_2: {
	struct byte_string_object* str;
	int size;
	int written;
	CheckOutput();
	GCSET_MESSAGE_STR_ARG(str, message, 0);
	size = generic_string_size(str);
#ifdef DIST
	if( g_self->outfile == stdout ){
	  UserWrite(generic_string_body(str), size);
	  written = size;
	}else if( g_self->outfile == stderr ){
	  UserEWrite(generic_string_body(str), size);
	  written = size;
	}else
#endif
	{
	  written = klic_fwrite(generic_string_body(str),
				1, size, g_self->outfile);
	}
	G_UNIFY_VALUE(arg(message, 1), makeint(written));
	break;
      }

      case functor_fwrite_1: {
	struct byte_string_object* str;
	char* strbody;
	int size;
	int written = 0;
	CheckOutput();
	GCSET_MESSAGE_STR_ARG(str, message, 0);
	strbody = (char*) generic_string_body(str);
	size = generic_string_size(str);
#ifdef DIST
	if( g_self->outfile == stdout ){
	  UserWrite(strbody, size);
	}else if( g_self->outfile == stderr ){
	  UserEWrite(strbody, size);
	}else
#endif
	{
	  while( size > 0 ){
	    written =
	      klic_fwrite(strbody+written, 1, size, g_self->outfile);
	    size -= written;
	  }
	}
	break;
      }

	/*** Common Messages ***/

      case functor_feof_1: {
	long iseof;
#ifdef DIST
	if( g_self->infile == stdin ||
	    g_self->outfile == stdout ||
	    g_self->outfile == stderr ){ iseof = 0; }
	else
#endif
	{
	  if( g_self->infile != NULL ){
	    iseof = feof(g_self->infile);
	  }else{
	    iseof = feof(g_self->outfile);
	  }
	}
	G_UNIFY_VALUE(arg(message, 0), makeint(iseof));
	break;
      }

      case functor_fseek_3: {
	long offset, whence, result;
	GCSET_MESSAGE_INT_ARG(offset, message, 0);
	GCSET_MESSAGE_INT_ARG(whence, message, 1);
#ifdef DIST
	if( g_self->infile == stdin ||
	    g_self->outfile == stdout || g_self->outfile == stderr ){
	  result = 0;
	}else
#endif
	{
	  FILE* file =
	    (g_self->infile == NULL ? g_self->outfile : g_self->infile);
	  result = fseek(file, offset, whence);
	}
	G_UNIFY_VALUE(arg(message, 2), makeint(result));
	break;
      }

      case functor_ftell_1: {
	long result;
#ifdef DIST
	if( g_self->infile == stdin ||
	    g_self->outfile == stdout || g_self->outfile == stderr ){
	  result = 0;
	}else
#endif
	{
	  FILE* file =
	    (g_self->infile == NULL ? g_self->outfile : g_self->infile);
	  result = ftell(file);
	}
	G_UNIFY_VALUE(arg(message, 0), makeint(result));
	break;
      }

      case functor_fflush_1: {
	long result = do_flush(g_self);
	G_UNIFY_VALUE(arg(message, 0), makeint(result));
	break;
      }

      case functor_fclose_1: {
	long result;
#ifdef DIST
	if( g_self->infile == stdin ||
	    g_self->outfile == stdout || g_self->outfile == stderr ){
	  result = 0;
	}else
#endif
	{
	  if( g_self->outfile != NULL &&
	      g_self->outfile != stdout &&
	      g_self->outfile != stderr ){
	    result = fclose(g_self->outfile);
	    g_self->outfile = NULL;
	  }else{
	    result = 0;
	  }
	  if( result == 0 ){
	    if( g_self->infile != NULL &&
		g_self->infile != stdin ){
	      result = fclose(g_self->infile);
	      g_self->infile = NULL;
	    }
	  }
	}
	G_UNIFY_VALUE(arg(message, 0), makeint(result));
	break;
      }

      case functor_sync_1: {
	G_UNIFY_VALUE(arg(message, 0), makeint(0));
	break;
      }

      default: goto message_error;
      }
    }else{
      goto message_error;
    }
    g_term = cdr_of(g_term);
    goto top;

  message_error:
    {
      char* inname = (char*) generic_string_body(functorp(g_self->inname));
      char* outname = (char*) generic_string_body(functorp(g_self->outname));
      if( inname == NULL ) inname = "";
      if( outname == NULL ) outname = "";
      debug_printf("Invalid message %k to I/O stream for %s%s%s\n",
		   message,
		   inname,
		   ((*inname != '\0' && *outname !='\0') ? "/" : ""),
		   outname);
      fatal("message error");
    }
  }else if( g_term==NILATOM ){
#ifdef DIST
    if( g_self->outfile == stdout ){
      UserFflush();
    }else if( g_self->outfile == stderr ){
      UserEFflush();
    }else if( g_self->infile == stdin ){
    }else
#endif
    {
      if( g_self->infile != NULL && g_self->infile != stdin ){
	fclose(g_self->infile);
      }
      if( g_self->outfile != NULL &&
	  g_self->outfile != stdout &&
	  g_self->outfile != stderr ){
	fclose(g_self->outfile);
      }
    }
    GC_TERMINATE;
  }else if( isref(g_term) ){
    q temp = derefone(g_term);
    if( isref(temp) &&
	(temp == g_term || derefone(temp) == g_term) ){
      reason = g_term;
      g_self->stream = 0;
      goto suspend;
    }else{
      g_term = temp;
      goto top;
    }
  }

 gc_request:
  G_MAKE_VAR(newvar);
  G_KL1_UNIFY(g_term, newvar);
  GC_RETURN_WITH_HOOK(newvar);

 suspend:
  GC_RETURN_WITH_HOOK(reason);
}

G_DEF_GC()
{
  G_STD_DECL;
  G_OBJ_TYPE* newself;

  G_SET_NEWOBJ_IN_NEWGEN(newself);
  *newself = *g_self;
  if( g_self->stream != 0 ){
    G_COPY_KL1_TERM_TO_NEWGEN(g_self->stream, newself->stream);
  }else{
    newself->stream = 0;
  }
  G_COPY_KL1_TERM_TO_NEWGEN(g_self->inname, newself->inname);
  G_COPY_KL1_TERM_TO_NEWGEN(g_self->outname, newself->outname);
  G_RETURN_FROM_GC(newself);
}

G_DEF_PRINT()
{
  G_STD_DECL;
  G_PRINT("$$FILE I/O$");
  G_RETURN_FROM_PRINT();
}

#define GCUSE_MY_UNIFY
#define GCUSE_MY_PRINT
#define GCUSE_MY_GC

/* define the method table structure of the merger */
#include <klic/gc_methtab.h>

G_DEF_NEW()
{
  G_STD_DECL_FOR_NEW();
  G_OBJ_TYPE* newobj;
  q infile, outfile, inpath, outpath;
  q var;
  struct pointer_object* inptr;
  struct pointer_object* outptr;

  /*
   * Arguments are:
   *   0: Input file object
   *   1: Its name
   *   2: Output file object
   *   3: Its name
   */
  if( g_argc != 4 ) G_ERROR_IN_NEW("Arity mismatch");

  infile = g_argv[0];
  G_DEREF_FOR_NEW(infile);
  if( infile != NILATOM &&
      (!G_ISGOBJ(infile) ||
       (struct data_object_method_table*) functor_of(infile) !=
       &pointer_g_method_table) ){
    fatal("First argument for file creation is not a pointer object");
  }
  inpath = g_argv[1];
  G_DEREF_FOR_NEW(inpath);

  outfile = g_argv[2];
  G_DEREF_FOR_NEW(outfile);
  if( outfile != NILATOM &&
      (!G_ISGOBJ(outfile) ||
       (struct data_object_method_table*) functor_of(outfile) !=
       &pointer_g_method_table) ){
    fatal("Third argument for file creation is not a pointer object");
  }
  outpath = g_argv[3];
  G_DEREF_FOR_NEW(outpath);

  G_SET_NEWOBJ_FOR_NEW(newobj, G_OBJ_SIZE);
  if( infile == NILATOM ){
    newobj->infile = NULL;
  }else{
    newobj->infile = (FILE*)
      ((struct pointer_object*) functorp(infile))->pointer;
  }
  if( outfile == NILATOM ){
    newobj->outfile = NULL;
  }else{
    newobj->outfile = (FILE*)
      ((struct pointer_object*) functorp(outfile))->pointer;
  }
  newobj->linecount = 0;
  newobj->inname = inpath;
  newobj->outname = outpath;
  newobj->stream = 0;
  var = GC_MAKE_HOOK_VAR((struct consumer_object*) newobj);
  return var;
}
