/* ---------------------------------------------------------- 
%   (C)1993,1994 Institute for New Generation Computer Technology 
%       (Read COPYRIGHT for detailed information.) 
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

#include <stdio.h>
#include <klic/gdobject.h>  /* debug_printf */
#include <klic/generic.h>  /* isclass */
#include "atom.h"
#include "funct.h"

struct merge_object {
  struct consumer_object_method_table *method_table;
  unsigned long count;
  q        outstream;
};

#define CLASS_NAME  merge
#define G_OBJ_TYPE  struct merge_object
#define G_OBJ_SIZE  G_SIZE_IN_Q(G_OBJ_TYPE)

#include <klic/gc_macro.h>
#include <klic/gd_macro.h>

#define GC_MERGE_IN_GOAL &predicate_gcmerge_xin_4

#define GC_MAKE_MERGE_IN_GOAL(newvar,vec,index,size) \
do{ \
  struct goalrec *goalp1; \
  goalp1 = (struct goalrec*) klic_alloc(G_SIZE_IN_Q(struct goalrec) - 2); \
  goalp1->pred = GC_MERGE_IN_GOAL; \
  goalp1->args[0] = newvar; \
  goalp1->args[1] = vec; \
  goalp1->args[2] = makeint(index); \
  goalp1->args[3] = makeint(size); \
  resume_same_prio(goalp1); \
}while(0)

G_USE_PREDICATE(predicate_gcmerge_xin_4);

GD_USE_CLASS(vector);

/* basic method definitions */

GCDEF_UNIFY()
{
  G_STD_DECL;
  q newvar;

 top:
  G_SWITCH_ON_TERM(list0, atomic0, functor0, generic_data0, susp0);

  list0:
    {
      struct cons* newout;
      q res;

      G_HEAPALLOC_WITH_CHECK(newout, 2, (struct cons*), res);
      if( res == GENERIC_GCREQUEST ){
	G_MAKE_VAR(newvar);
	G_KL1_UNIFY(g_term, newvar);
	GC_RETURN_WITH_HOOK(newvar);
      }
      newout->car = car_of(g_term);
      newout->cdr = makeref(&(newout->cdr));
      do_shallow_unify(g_self->outstream, makecons(newout));
      g_self->outstream = newout->cdr;
      g_term = cdr_of(g_term);
      goto top;
    }
  atomic0:
    if (g_term == NILATOM) {
      if (--(g_self->count) == 0) {
	do_shallow_unify(g_self->outstream, NILATOM);
      }
      GC_TERMINATE;
    } else goto invalid_data;
  functor0:
    goto invalid_data;
  generic_data0:
    if( isclass(g_term, vector) ){
      unsigned long size;
      unsigned long i;
      {
	q argv[1];
	GD_CALL_GMETHOD(g_term, vector_1, argv);
	size = intval(argv[0]);
      }
      g_self->count += size - 1;
      if (g_self->count == 0) {
	do_shallow_unify(g_self->outstream, NILATOM);
      } else {
	q argv[2];
	q hook_var;
	for (i=0; i<size; i++) {
	  q res;

	  argv[0] = makeint(i);
	  GD_CALL_GMETHOD(g_term, element_2, argv);
	  G_HEAPALLOC_WITH_CHECK(hook_var, 1, (q), res);
	  if( res == GENERIC_GCREQUEST ){
	    g_self->count -= size - i -1;
	    G_MAKE_VAR(newvar);
	    GC_MAKE_MERGE_IN_GOAL(newvar, g_term, i, size);
	    GC_RETURN_WITH_HOOK(newvar);
	  }
	  derefone(hook_var) =
	    GC_MAKE_HOOK_VAR((struct consumer_object*) g_self);
	  G_KL1_UNIFY(argv[1], hook_var);
	}
      }
      GC_TERMINATE;
    } else goto invalid_data;
 susp0: 
    GC_RETURN_WITH_HOOK(g_term);
 invalid_data:
  debug_printf("### %k ###\n", g_term);
  fatal("Invalid data unified with merger");
}

G_DEF_GC()
{
  G_STD_DECL;
  G_OBJ_TYPE* newself;

  G_SET_NEWOBJ_IN_NEWGEN(newself);
  newself->count = g_self->count;
  G_COPY_KL1_TERM_TO_NEWGEN(g_self->outstream, newself->outstream);
  G_RETURN_FROM_GC(newself);
}


G_DEF_PRINT()
{
  G_STD_DECL;
  G_PRINT("$$MERGER$");
  G_RETURN_FROM_PRINT();
}

#define GCUSE_MY_UNIFY
#define GCUSE_MY_PRINT
#define GCUSE_MY_GC

/* define the method table structure of the merger */
#include <klic/gc_methtab.h>

/*  new_vector function */
G_DEF_NEW()
{
  G_STD_DECL_FOR_NEW();
  G_OBJ_TYPE* newmerger;
  q var;

  if (g_argc != 1) G_ERROR_IN_NEW("Arity mismatch");

  G_SET_NEWOBJ_FOR_NEW(newmerger, G_OBJ_SIZE);
  newmerger->count = 1;
  newmerger->outstream = g_argv[0];
  var = GC_MAKE_HOOK_VAR((struct consumer_object*) newmerger);
  return var;
}
