/* ---------------------------------------------------------- 
%   (C)1993,1994,1995 Institute for New Generation Computer Technology 
%       (Read COPYRIGHT for detailed information.) 
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

/*
  Vector Objects
*/

#include <stdio.h>
#include <klic/gdobject.h>  /* fatal, debug_printf */
#include <klic/g_vector.h>

#ifdef DIST
#include "interpe.h"
#endif

#include "atom.h"
#include "funct.h"

#define CLASS_NAME  vector
#define G_OBJ_TYPE  struct vector_object
#define G_OBJ_SIZE  (G_SIZE_IN_Q(G_OBJ_TYPE))

#include <klic/gd_macro.h>

#define VECTOR_OBJ(x)	((G_OBJ_TYPE*) functorp(x))
#define Shallow(vector) \
do{ if( !IS_SHALLOW_VECTOR(vector) ) do_shallow(vector); }while(0)

extern q hash_kl1_term();

#ifdef DIST
extern void encode_data();
extern void push_decode_stack();
extern q pop_decode_stack();
extern q element_of_vector(q v, q k);
static void decode_vector(combuf* inbuf);
#endif /* DIST */


/* shallowing */

static void
do_shallow(vector)
     struct vector_object *vector;
{
  q v, last, next;

  /* Go down to the shallwed version inverting pointers */
  last = VECTOR_SHALLOW_MARK;
  v = makefunctor(vector);
  do {
    next = VECTOR_OBJ(v)->next;
    VECTOR_OBJ(v)->next = last;
    last = v;
    v = next;
  } while (v != VECTOR_SHALLOW_MARK);
  v = last;

  /* Update physical vector elements tracing the inverted pointers */
  {
    long size = VECTOR_OBJ(v)->index;
    q *body = VECTOR_OBJ(v)->body;
    next = VECTOR_OBJ(v)->next;
    do {
      long index;
      index = VECTOR_OBJ(next)->index;
      VECTOR_OBJ(v)->index = index;
      VECTOR_OBJ(v)->body = (q *)body[index];
      body[index] = (q)(VECTOR_OBJ(next)->body);
      v = next;
      next = VECTOR_OBJ(v)->next;
    } while (next != VECTOR_SHALLOW_MARK);
    VECTOR_OBJ(v)->index = size;
    VECTOR_OBJ(v)->body = body;
  }
}


/* basic method definitions */

GDDEF_GUNIFY()
{
  G_STD_DECL;
  long size, k;
  if( g_self->method_table != GD_OTHER->method_table ) return GENERIC_FAILED;
  Shallow(g_self);
  size = g_self->index;
  Shallow(GD_OTHER);
  if( GD_OTHER->index != size ) return GENERIC_FAILED;
  for (k=0; k<size; k++) {
    q retval;
    q x = GD_OTHER->body[k];
    Shallow(g_self);
    retval = eq_terms_body(g_self->body[k], x);
    switch ((long)retval) {
    case GENERIC_SUCCEEDED:
      break;
    case GENERIC_FAILED:
      return GENERIC_FAILED;
    default:
      return retval;
    }
    Shallow(GD_OTHER);
  }
  return GENERIC_SUCCEEDED;
}

GDDEF_UNIFY()
{
  G_STD_DECL;
  long size, k;

  if( g_self->method_table != GD_OTHER->method_table ) GD_UNIFY_FAIL;
  Shallow(g_self);
  size = g_self->index;
  Shallow(GD_OTHER);
  if (GD_OTHER->index != size) GD_UNIFY_FAIL;
  for (k=0; k<size; k++) {
    q x = GD_OTHER->body[k];
    Shallow(g_self);
    do_shallow_unify(g_self->body[k], x);
    Shallow(GD_OTHER);
  }
}

G_DEF_GC()
{
  G_STD_DECL;
  G_OBJ_TYPE* newself;

  if( g_self->iscnst ){
    fatal("GC method for constant vector unexpected");
  }
  G_SET_NEWOBJ_IN_NEWGEN(newself);
  newself->iscnst = 0;
  if( IS_SHALLOW_VECTOR(g_self) ){
    /* Shallow version */
    q* body = g_self->body;
    long size = g_self->index;
    q *newbody;
    long k;
    if(heapp() + size > real_heaplimit()) fatal("Not enough space collected");
    newbody = klic_alloc(size);
    newself->next = VECTOR_SHALLOW_MARK;
    newself->index = size;
    newself->body = newbody;
    for (k=0; k<size; k++) {
      G_COPY_KL1_TERM_TO_NEWGEN(body[k], newbody[k]);
    }
  } else {
    /* Deep version */
    newself->index = g_self->index;
    G_COPY_KL1_TERM_TO_NEWGEN(g_self->next, newself->next);

    /* G_COPY_KL1_TERM_TO_NEWGEN(g_self->body, newself->body); */
    newself->body = g_self->body;
    push_gc_stack((q*) &(newself->body));
  }
  G_RETURN_FROM_GC(newself);
}

#ifdef DIST
G_DEF_ENCODE()
{
  G_STD_DECL;
  int i;
  q   elem;
  long size;

  Shallow(g_self);
  size = g_self->index;
  depth--;
  for(i = 0; i<size; i++){
    elem = element_of_vector(makefunctor(g_self), makeint(i));
    if (elem == 0){
      fatal("Error in element_of_vector");
    }
    encode_data(buffer, elem, depth);
  }
  PUT_BUFFER(buffer, decode_vector);
  PUT_BUFFER(buffer, size);

  return GENERIC_SUCCEEDED;
}
#endif /* DIST */

/* Generic method */

GDDEF_METHOD(element_2)
{
  G_STD_DECL;
  int position;
  Shallow(g_self);
  GDSET_INTARG_WITHIN_RANGE(position, g_argv[0], 0, g_self->index);
  do_shallow_unify(g_argv[1], g_self->body[position]);
}

GDDEF_METHOD(size_1)
{
  G_STD_DECL;
  Shallow(g_self);
  G_UNIFY_VALUE(g_argv[0], makeint(g_self->index));
}

GDDEF_METHOD(set__element_3)
{
  G_STD_DECL;
  long position;
  long size;
  G_OBJ_TYPE* newvect;
  q *body;
  int iscnst;

  Shallow(g_self);
  size = g_self->index;
  iscnst = g_self->iscnst;
  body = g_self->body;
  GDSET_INTARG_WITHIN_RANGE(position, g_argv[0], 0, size);
  GDSET_NEWOBJ(newvect);
  if (!iscnst) {
    q olddata = body[position];

    g_self->index = position;
    g_self->body = (q*) olddata;
    body[position] = g_argv[1];
    g_self->next = makefunctor(newvect);
    newvect->body = body;
  } else {
    long k;
    q *newbody;
    GD_ALLOC_AREA(newbody,(q*),size);
    for (k=0; k<size; k++) newbody[k] = body[k];
    newbody[position] = g_argv[1];
    newvect->body = newbody;
  }
  newvect->next = VECTOR_SHALLOW_MARK;
  newvect->index = size;
  newvect->iscnst = 0;
  G_UNIFY_VALUE(g_argv[2], makefunctor(newvect));
}

GDDEF_METHOD(set__element_4)
{
  G_STD_DECL;
  long position;
  long size;
  G_OBJ_TYPE* newvect;
  q *body;
  int iscnst;
  q olddata;

  Shallow(g_self);
  size = g_self->index;
  body = g_self->body;
  iscnst = g_self->iscnst;

  GDSET_INTARG_WITHIN_RANGE(position, g_argv[0], 0, size);
  GDSET_NEWOBJ(newvect);
  olddata = body[position];
  do_shallow_unify(g_argv[1], olddata);
  if (!iscnst) {
    g_self->index = position;
    g_self->body = (q*) olddata;
    body[position] = g_argv[2];
    g_self->next = makefunctor(newvect);
    newvect->body = body;
  } else {
    long k;
    q *newbody;
    GD_ALLOC_AREA(newbody,(q*),size);
    for (k=0; k<size; k++) newbody[k] = body[k];
    newbody[position] = g_argv[2];
    newvect->body = newbody;
  }
  newvect->next = VECTOR_SHALLOW_MARK;
  newvect->index = size;
  newvect->iscnst = 0;
  G_UNIFY_VALUE(g_argv[3], makefunctor(newvect));
}

GDDEF_METHOD(split_3)
{
  G_STD_DECL;
  long split_point;
  unsigned long size, lower_size;
  q *upper_body, *lower_body;
  struct vector_object *upper, *lower;

  Shallow(g_self);
  size = g_self->index;
  GDSET_INTARG_WITHIN_RANGE(split_point, g_argv[0], 0, size + 1);
  lower_size = size-split_point;
  GD_ALLOC_AREA(upper_body, (q *), split_point);
  GD_ALLOC_AREA(lower_body, (q *), lower_size);
  GD_ALLOC_AREA(upper, (G_OBJ_TYPE*), G_OBJ_SIZE);
  GD_ALLOC_AREA(lower, (G_OBJ_TYPE*), G_OBJ_SIZE);
  BCOPY((char*) g_self->body, (char*) upper_body, split_point*sizeof(q));
  BCOPY((char*) (g_self->body + split_point), (char*) lower_body,
	lower_size*sizeof(q));
  upper->method_table = lower->method_table = g_self->method_table;
  upper->next = lower->next = VECTOR_SHALLOW_MARK;
  upper->index = split_point;
  lower->index = lower_size;
  upper->body = upper_body;
  lower->body = lower_body;
  upper->iscnst = lower->iscnst = 0;
  G_UNIFY_VALUE(g_argv[1], makefunctor(upper));
  G_UNIFY_VALUE(g_argv[2], makefunctor(lower));
}

GDDEF_METHOD(join_2)
{
  G_STD_DECL;
  unsigned long size1, size2, newsize;
  q anotherq = g_argv[0];
  G_OBJ_TYPE *another, *new;
  q *newbody;

  Shallow(g_self);
  size1 = g_self->index;
  GD_DEREF(anotherq);
  another = (G_OBJ_TYPE*) functorp(anotherq);
  if( !isfunctor(anotherq) ||
      another->method_table != g_self->method_table) {
    debug_printf("### %k ###\n", another);
    fatal("Invalid argument to string join");
  }
  Shallow(another);
  size2 = another->index;
  newsize = size1+size2;
  GD_ALLOC_AREA(newbody, (q *), newsize);
  GD_ALLOC_AREA(new, (G_OBJ_TYPE*), G_OBJ_SIZE);
  Shallow(g_self);
  BCOPY((char*) g_self->body, (char*) newbody, size1*sizeof(q));
  Shallow(another);
  BCOPY((char*) another->body, (char*) (newbody+size1), size2*sizeof(q));
  new->method_table = g_self->method_table;
  new->next = VECTOR_SHALLOW_MARK;
  new->index = newsize;
  new->body = newbody;
  new->iscnst = 0;
  G_UNIFY_VALUE(g_argv[1], makefunctor(new));
}

/*  Generic Method Table */
GDDEF_GENERIC()
{
  G_STD_DECL;

  switch( g_method_functor ){
    GD_METHOD_CASE(element_2);
    GD_METHOD_CASE(size_1);
    GD_METHOD_CASE(set__element_3);
    GD_METHOD_CASE(set__element_4);
    GD_METHOD_CASE(split_3);
    GD_METHOD_CASE(join_2);
    GD_METHOD_CASE_DEFAULT;
  }
}

/* guard generic methods */

GDDEF_GMETHOD(element_2)
{
  G_STD_DECL;
  long position;

  Shallow(g_self);
  GDSET_GINTARG_WITHIN_RANGE(position, g_argv[0], 0, g_self->index);
  g_argv[1] = g_self->body[position];
  return GENERIC_SUCCEEDED;
}

GDDEF_GMETHOD(vector_1)
{
  G_STD_DECL;
  Shallow(g_self);
  g_argv[0] = makeint(g_self->index);
  return GENERIC_SUCCEEDED;
}

GDDEF_GGENERIC()
{
  G_STD_DECL;
  switch( g_method_functor ){
    GD_GMETHOD_CASE(element_2);
    GD_GMETHOD_CASE(vector_1);
    GD_GMETHOD_CASE_DEFAULT;
  }
}

G_DEF_PRINT()
{
  G_STD_DECL;
  unsigned long size, limit, k;
  Shallow(g_self);
  size = g_self->index;
  limit = (size > g_length ? g_length : size);
  G_PRINT("{");
  if( size != 0 ) GD_PRINT_KL1_TERM(g_self->body[0]);
  for (k=1; k<limit; k++) {
    G_PRINT(",");
    if( size>0 ) GD_PRINT_KL1_TERM(g_self->body[k]);
  }
  if (size != limit) G_PRINT(",..");
  G_PRINT("}");
  G_RETURN_FROM_PRINT();
}

GDDEF_COMPARE()
{
  G_STD_DECL;
  unsigned long size1, size2, limit, k;
  Shallow(g_self);
  size1 = g_self->index;
  Shallow(GD_OTHER);
  size2 = GD_OTHER->index;
  limit = (size1 <= size2 ? size1 : size2);
  for (k = 0; k < limit; k++) {
    q elem, tmp;
    Shallow(g_self);
    elem = g_self->body[k];
    Shallow(GD_OTHER);
    tmp = builtin_compare_terms(elem, GD_OTHER->body[k]);
    if( !isint(tmp) ) return tmp;
    if( tmp != makeint(0) ) return tmp;
  }
  if (size1 != size2) {
    return makeint(size1 >= size2 ? size1+1 : -(size2+1));
  } else {
    return makeint(0);
  }
}

GDDEF_HASH()
{
  G_STD_DECL;
  unsigned long size;
  Shallow(g_self);
  size = g_self->index;
  if (GD_LEVEL == 0 || size == 0) {
    return makeint((long) g_self->method_table + 0xb5d*size);
  } else {
    q h1, h2, h3;
    GD_LEVEL--;
    h1 = hash_kl1_term(g_self->body[0], GD_LEVEL);
    if (!isint(h1)) return h1;
    h2 = hash_kl1_term(g_self->body[size>>1], GD_LEVEL);
    if (!isint(h1)) return h2;
    h3 = hash_kl1_term(g_self->body[size-1], GD_LEVEL);
    if (!isint(h1)) return h3;
    return
      makeint(0x813*intval(h1) + 0x425*intval(h2) + 0x3c9*intval(h3) +
		(long) g_self->method_table + 0xb5d*size );
  }
}

#define GDUSE_MY_GUNIFY
#define GDUSE_MY_UNIFY
#define GDUSE_MY_PRINT
#define GDUSE_MY_GC
#ifdef DIST
#define GDUSE_MY_ENCODE
#endif
#define GDUSE_MY_GENERIC
#define GDUSE_MY_GGENERIC
#define GDUSE_MY_COMPARE
#define GDUSE_MY_HASH

/* define the method table structure of the vector */

#include <klic/gd_methtab.h>

/*  new_vector function */
G_DEF_NEW()
{
  G_STD_DECL_FOR_NEW();
  G_OBJ_TYPE* newvect;
  q *body;
  long size, k;
  q init;

  if( g_argc != 1 ) G_ERROR_IN_NEW("Too few or too many arguments");
  init = g_argv[0];
  G_DEREF_FOR_NEW(init);
  G_SET_NEWOBJ_FOR_NEW(newvect, G_SIZE_IN_Q(G_OBJ_TYPE));
  if( isint(init) ){
    size = intval(init);
    if (size < 0) G_ERROR_IN_NEW("Negative size specified");
    GD_ALLOC_AREA_FOR_NEW(body, (q*), size);
    for (k=0; k<size; k++) body[k] = makeint(0L);
  }else if( init==NILATOM || iscons(init) ){
    for (size=0; ; size++) {
      if (init == NILATOM) break;
      init = cdr_of(init);
      G_DEREF_FOR_NEW(init);
      if( init != NILATOM && !iscons(init) )
	G_ERROR_IN_NEW("Invalid parameter");
    }
    init = g_argv[0];
    GD_ALLOC_AREA_FOR_NEW(body, (q*), size);
    for (k=0; k<size; k++) {
      G_DEREF_FOR_NEW(init);
      body[k] = car_of(init);
      init = cdr_of(init);
    }
  } else {
    G_ERROR_IN_NEW("Invalid parameter");
  }
  newvect->next = VECTOR_SHALLOW_MARK;
  newvect->iscnst = 0;
  newvect->index = size;
  newvect->body = body;
  return makefunctor(newvect);
}

/* Interface with builtin predicates */

extern q
create_vector(q* body, long size)
{
  G_STD_DECL;
  struct vector_object* newvect;
  heapalloc(newvect, G_SIZE_IN_Q(G_OBJ_TYPE), (G_OBJ_TYPE*));
  newvect->method_table = &G_method_table;
  newvect->next = VECTOR_SHALLOW_MARK;
  newvect->iscnst = 0;
  newvect->index = size;
  newvect->body = body;
  return makefunctor(newvect);
}

extern q
size_of_vector(v)
     q v;
{
  Shallow(VECTOR_OBJ(v));
  return makeint(VECTOR_OBJ(v)->index);
}

extern q
element_of_vector(v, k)
     q v, k;
{
  Shallow(VECTOR_OBJ(v));
  if( intval(k) < 0 || VECTOR_OBJ(v)->index <= intval(k) ){
    return (q) 0;
  } else {
    return VECTOR_OBJ(v)->body[intval(k)];
  }
}

#ifdef DIST
static void
decode_vector(inbuf)
     combuf *inbuf;
{
  G_STD_DECL;
  G_OBJ_TYPE* newvect;
  q *body;
  long size, k;
  q res;

  G_HEAPALLOC_WITH_CHECK(newvect, G_SIZE_IN_Q(G_OBJ_TYPE),
                          (G_OBJ_TYPE*), res);
  if(res == GENERIC_GCREQUEST){
      add_this_more_space(G_SIZE_IN_Q(G_OBJ_TYPE));
      inbuf->rd_index--;
      return;
  }
  newvect->method_table = &G_method_table;
  newvect-> index = size = (long)GET_BUFFER(inbuf);

  G_HEAPALLOC_WITH_CHECK(body, size, (q*), res); 
  if(res == GENERIC_GCREQUEST){
      add_this_more_space(size);
      inbuf->rd_index -= 2;
      return;
  }

  for(k = size; k > 0 ; k--){
    body[k-1] = pop_decode_stack();
  }

  newvect->next   = VECTOR_SHALLOW_MARK;
  newvect->iscnst = 0;
  newvect->body = body;

  push_decode_stack((q)makefunctor(newvect));
}
#endif /* DIST */
