/* ---------------------------------------------------------- 
%   (C)1993,1994,1995 Institute for New Generation Computer Technology 
%       (Read COPYRIGHT for detailed information.) 
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

#include <stdio.h>
#include <ctype.h>
#include <klic/gdobject.h>  /* malloc_check, fatal, debug_printf */
#include <klic/g_string.h>
#include "atom.h"
#include "funct.h"

#ifdef DIST
#include "interpe.h"
#endif

#define CLASS_NAME  byte__string
#define G_OBJ_TYPE  struct byte_string_object
#define G_OBJ_SIZE  (G_SIZE_IN_Q(G_OBJ_TYPE))

#include <klic/gd_macro.h>

#define ELEMSIZE 8
#define STRING_OBJ(x)	((G_OBJ_TYPE*) functorp(x))
#define ROUND_UP(size)	(((size)+sizeof(q)-1)/sizeof(q))
#define Shallow(string) \
do{ if( !IS_SHALLOW_STRING(string) ) do_shallow(string); }while(0)

#ifdef DIST
extern void push_decode_stack();
extern unsigned char* generic_string_body(G_OBJ_TYPE* str);
static void decode_byte_string(combuf* inbuf);
#endif /* DIST */


/* shallowing */

static void
do_shallow(string)
     struct byte_string_object *string;
{
  q s, last, next;

  /* Go down to the shallwed version inverting pointers */
  last = STRING_SHALLOW_MARK;
  s = makefunctor(string);
  do {
    next = STRING_OBJ(s)->next;
    STRING_OBJ(s)->next = last;
    last = s;
    s = next;
  } while (s != STRING_SHALLOW_MARK);
  s = last;

  /* Update physical string elements tracing the inverted pointers */
  {
    long size = STRING_OBJ(s)->index;
    unsigned char *body = STRING_OBJ(s)->body;
    next = STRING_OBJ(s)->next;
    do {
      long index;
      index = STRING_OBJ(next)->index;
      STRING_OBJ(s)->index = index;
      STRING_OBJ(s)->body = (unsigned char *)((unsigned long)body[index]);
      body[index] = (unsigned char)(unsigned long)(STRING_OBJ(next)->body);
      s = next;
      next = STRING_OBJ(s)->next;
    } while (next != STRING_SHALLOW_MARK);
    STRING_OBJ(s)->index = size;
    STRING_OBJ(s)->body = body;
  }
}


/* basic method definitions */

GDDEF_GUNIFY()
{
  G_STD_DECL;
  long size, k;
  if( g_self->method_table != GD_OTHER->method_table ) return GENERIC_FAILED;
  Shallow(g_self);
  size = g_self->index;
  Shallow(GD_OTHER);
  if( GD_OTHER->index != size ) return GENERIC_FAILED;
  if( IS_SHALLOW_STRING(g_self) ){
    return (BCMP(g_self->body, GD_OTHER->body, size) == 0 ?
	    GENERIC_SUCCEEDED : GENERIC_FAILED );
  }
  for( k=0; k<size; k++ ){
    unsigned char c = GD_OTHER->body[k];
    Shallow(g_self);
    if( g_self->body[k] != c ) return GENERIC_FAILED;
    Shallow(GD_OTHER);
  }
  return GENERIC_SUCCEEDED;
}

GDDEF_UNIFY()
{
  G_STD_DECL;
  long size, k;
  if (g_self->method_table != GD_OTHER->method_table) GD_UNIFY_FAIL;
  Shallow(g_self);
  size = g_self->index;
  Shallow(GD_OTHER);
  if (GD_OTHER->index != size) GD_UNIFY_FAIL;
  if (IS_SHALLOW_STRING(g_self)) {
    if (BCMP(g_self->body, GD_OTHER->body, size) != 0) GD_UNIFY_FAIL;
  } else {
    for (k=0; k<size; k++) {
      unsigned char c = GD_OTHER->body[k];
      Shallow(g_self);
      if (g_self->body[k] != c) GD_UNIFY_FAIL;
      Shallow(GD_OTHER);
    }
  }
}

G_DEF_GC()
{
  G_STD_DECL;
  G_OBJ_TYPE* newself;

  G_SET_NEWOBJ_IN_NEWGEN(newself);
  if (g_self->iscnst) {
    fatal("GC method for constant string unexpected");
  }
  newself->iscnst = 0;
  if (IS_SHALLOW_STRING(g_self)) {
    unsigned char *body = g_self->body;
    unsigned char *newbody;
    unsigned long size = g_self->index;
    unsigned long qsize = ROUND_UP(size);
    if(heapp() + qsize > real_heaplimit()) fatal("Not enough space collected");
    newbody = (unsigned char*) klic_alloc(qsize);
    newself->next = STRING_SHALLOW_MARK;
    newself->index = size;
    newself->body = newbody;
    BCOPY(body,newbody,qsize*sizeof(q));
  } else {
    newself->index = g_self->index;
    newself->body = g_self->body;
    G_COPY_KL1_TERM_TO_NEWGEN(g_self->next, newself->next);
  }
  G_RETURN_FROM_GC(newself);
}

#ifdef DIST
/* for byte string */
G_DEF_ENCODE()  
{
  G_STD_DECL;
  long size;
  int i;
  long *tmp_buf;

  Shallow(g_self);

  PUT_BUFFER(buffer, decode_byte_string);
  PUT_BUFFER(buffer, g_self->index);
  size = ROUND_UP(g_self->index);
  
  tmp_buf = (long *)generic_string_body(g_self);
  for(i = 0 ; i < size ; i++){
    PUT_BUFFER(buffer, tmp_buf[i]);  /* Should be carefule of endian */
  }
  return GENERIC_SUCCEEDED;
}
#endif /* DIST */

/* Generic method */

GDDEF_METHOD(string_2)
{
  G_STD_DECL;
  Shallow(g_self);
  G_UNIFY_VALUE(g_argv[0], makeint(g_self->index));
  G_UNIFY_VALUE(g_argv[1], makeint(ELEMSIZE));
}

GDDEF_METHOD(element_2)
{
  G_STD_DECL;
  long position;
  Shallow(g_self);
  GDSET_INTARG_WITHIN_RANGE(position, g_argv[0], 0, (long) g_self->index);
  G_UNIFY_VALUE(g_argv[1], makeint((unsigned long) g_self->body[position]));
}

GDDEF_METHOD(size_1)
{
  G_STD_DECL;
  Shallow(g_self);
  G_UNIFY_VALUE(g_argv[0], makeint(g_self->index));
}

GDDEF_METHOD(element__size_1)
{
  G_STD_DECL;
  Shallow(g_self);
  G_UNIFY_VALUE(g_argv[0], makeint(ELEMSIZE));
}

GDDEF_METHOD(set__element_3)
{
  G_STD_DECL;
  long position, newelem;
  G_OBJ_TYPE* newstr;
  unsigned char *body;
  int iscnst;
  long size;

  Shallow(g_self);
  size = g_self->index;
  iscnst = g_self->iscnst;
  body = g_self->body;
  GDSET_INTARG_WITHIN_RANGE(position, g_argv[0], 0, size);
  GDSET_INTARG_WITHIN_RANGE(newelem, g_argv[1], 0, 256);
  GDSET_NEWOBJ(newstr);
  if (!iscnst) {
    unsigned char olddata = body[position];

    g_self->index = position;
    g_self->body = (unsigned char *)(unsigned long)olddata;
    body[position] = newelem;
    g_self->next = makefunctor(newstr);
    newstr->body = body;
  } else {
    unsigned char *newbody;
    unsigned long qsize = ROUND_UP(size);
    GD_ALLOC_AREA(newbody, (unsigned char *), qsize);
    BCOPY(body, newbody, qsize*sizeof(q));
    newbody[position] = newelem;
    newstr->body = newbody;
  }
  newstr->next = STRING_SHALLOW_MARK;
  newstr->index = size;
  newstr->iscnst = 0;
  G_UNIFY_VALUE(g_argv[2], makefunctor(newstr));
}

GDDEF_METHOD(split_3)
{
  G_STD_DECL;
  unsigned long size, lower_size;
  long split_point;
  unsigned char *upper_body, *lower_body;
  struct byte_string_object *upper, *lower;

  Shallow(g_self);
  size = g_self->index;
  GDSET_INTARG_WITHIN_RANGE(split_point, g_argv[0], 0, size + 1);
  lower_size = size-split_point;
  GD_ALLOC_AREA(upper_body, (unsigned char *), ROUND_UP(split_point));
  GD_ALLOC_AREA(lower_body, (unsigned char *), ROUND_UP(lower_size));
  GD_ALLOC_AREA(upper, (G_OBJ_TYPE*), G_OBJ_SIZE);
  GD_ALLOC_AREA(lower, (G_OBJ_TYPE*), G_OBJ_SIZE);
  BCOPY(g_self->body, upper_body, split_point);
  BCOPY(g_self->body+split_point, lower_body, lower_size);
  upper->method_table = lower->method_table = g_self->method_table;
  upper->next = lower->next = STRING_SHALLOW_MARK;
  upper->index = split_point;
  lower->index = lower_size;
  upper->iscnst = lower->iscnst = 0;
  upper->ismbdiff = lower->ismbdiff = 0;
  upper->body = upper_body;
  lower->body = lower_body;
  G_UNIFY_VALUE(g_argv[1], makefunctor(upper));
  G_UNIFY_VALUE(g_argv[2], makefunctor(lower));
}

GDDEF_METHOD(join_2)
{
  G_STD_DECL;
  unsigned long size1, size2, newsize;
  q anotherq = g_argv[0];
  G_OBJ_TYPE *another, *new;
  unsigned char *newbody;

  Shallow(g_self);
  size1 = g_self->index;
  GD_DEREF(anotherq);
  another = (G_OBJ_TYPE*) functorp(anotherq);
  if( !isfunctor(anotherq) ||
      another->method_table != g_self->method_table) {
    debug_printf("### %k ###\n", another);
    fatal("Invalid argument to string join");
  }
  Shallow(another);
  size2 = another->index;
  newsize = size1+size2;
  GD_ALLOC_AREA(newbody, (unsigned char *), ROUND_UP(newsize));
  GD_ALLOC_AREA(new, (G_OBJ_TYPE*), G_OBJ_SIZE);
  Shallow(g_self);
  BCOPY(g_self->body, newbody, size1);
  Shallow(another);
  BCOPY(another->body, newbody+size1, size2);
  new->method_table = g_self->method_table;
  new->next = STRING_SHALLOW_MARK;
  new->index = newsize;
  new->iscnst = 0;
  new->ismbdiff = 0;
  new->body = newbody;
  G_UNIFY_VALUE(g_argv[1], makefunctor(new));
}

GDDEF_METHOD(search__character_4)
{
  G_STD_DECL;
  long start, end, code;
  unsigned long k;

  Shallow(g_self);
  GDSET_INTARG_WITHIN_RANGE(start, g_argv[0], 0, (long) g_self->index);
  GDSET_INTARG_WITHIN_RANGE(end,   g_argv[1], 0, (long) g_self->index);
  GDSET_INTARG_WITHIN_RANGE(code,  g_argv[2], 0, 256L);
  if (start <= end) {
    for (k=start; k<=end; k++) {
      if (((unsigned long) g_self->body[k]) == code) goto done;
    }
  } else {
    for (k=start; k>=end; k--) {
      if (((unsigned long) g_self->body[k]) == code) goto done;
    }
  }
  G_UNIFY_VALUE(g_argv[3], makeint(-1L));
  return;

 done:
  G_UNIFY_VALUE(g_argv[3], makeint(k));
}


/*  Generic Method Table */
GDDEF_GENERIC()
{
  G_STD_DECL;

  switch( g_method_functor ){
    GD_METHOD_CASE(string_2);
    GD_METHOD_CASE(element_2);
    GD_METHOD_CASE(size_1);
    GD_METHOD_CASE(element__size_1);
    GD_METHOD_CASE(set__element_3);
    GD_METHOD_CASE(split_3);
    GD_METHOD_CASE(join_2);
    GD_METHOD_CASE(search__character_4);
    GD_METHOD_CASE_DEFAULT;
  }
}

/* guard generic methods */

GDDEF_GMETHOD(element_2)
{
  G_STD_DECL;
  long position;

  Shallow(g_self);
  GDSET_GINTARG_WITHIN_RANGE(position, g_argv[0], 0, (long) g_self->index);
  g_argv[1] = makeint((long) g_self->body[position]);
  return GENERIC_SUCCEEDED;
}

GDDEF_GMETHOD(string_2)
{
  G_STD_DECL;

  Shallow(g_self);
  g_argv[0] = makeint(g_self->index);
  g_argv[1] = makeint(ELEMSIZE);
  return GENERIC_SUCCEEDED;
}

static int compare_two_strings(G_OBJ_TYPE* s1, G_OBJ_TYPE* s2)
{
  long size1, size2, minsize, k;
  Shallow(s1);
  size1 = s1->index;
  Shallow(s2);
  size2 = s2->index;
  minsize = ((size1 < size2) ? size1 : size2);
  if (!IS_SHALLOW_STRING(s1)) {
    /* s1 and s2 are different versions of the same string */
    for (k=0; k<minsize; k++) {
      unsigned char c;
      Shallow(s1);
      c = s1->body[k];
      Shallow(s2);
      if (c != s2->body[k]) {
	return ((c < s2->body[k]) ? -(k+1) : k+1);
      }
    }
  } else {
    for (k=0; k<minsize; k++) {
      if (s1->body[k] != s2->body[k]) {
	return ((s1->body[k] < s2->body[k]) ? -(k+1) : k+1);
      }
    }
  }
  if (size1 != size2) {
    return ((size1<size2) ? -(size1+1) : size2+1);
  } else {
    return 0;
  }
}

GDDEF_GMETHOD(less__than_1)
{
  G_STD_DECL;
  q otherq = g_argv[0];
  G_OBJ_TYPE* other;
  int cmp;

  if( !G_ISGOBJ(otherq) ) return GENERIC_FAILED;
  other = (G_OBJ_TYPE*) functorp(otherq);
  if( other->method_table != g_self->method_table ) return GENERIC_FAILED;
  cmp = compare_two_strings(g_self, other);
  return cmp < 0 ? GENERIC_SUCCEEDED : GENERIC_FAILED;
}

GDDEF_GMETHOD(not__less__than_1)
{
  G_STD_DECL;
  q otherq = g_argv[0];
  G_OBJ_TYPE* other;
  int cmp;

  if( !G_ISGOBJ(otherq) ) return GENERIC_FAILED;
  other = (G_OBJ_TYPE*) functorp(otherq);
  if( other->method_table != g_self->method_table ) return GENERIC_FAILED;
  cmp = compare_two_strings(g_self, other);
  return cmp >= 0 ? GENERIC_SUCCEEDED : GENERIC_FAILED;
}

GDDEF_GMETHOD(estring_3)
{
  G_STD_DECL;
  long size, k;
  q tmp;

  Shallow(g_self);
  size = g_self->index;
  if( intval(g_argv[0]) != size ) return GENERIC_FAILED;
  if( intval(g_argv[1]) != ELEMSIZE ) return GENERIC_FAILED;
  tmp = g_argv[2];
  for( k=0; k < size; k++, tmp = cdr_of(tmp) ){
    if( intval(car_of(tmp)) != g_self->body[k] ) return GENERIC_FAILED;
  }
  return GENERIC_SUCCEEDED;
}

GDDEF_GGENERIC()
{
  G_STD_DECL;

  switch( g_method_functor ){
    GD_GMETHOD_CASE(element_2);
    GD_GMETHOD_CASE(string_2);
    GD_GMETHOD_CASE(less__than_1);
    GD_GMETHOD_CASE(not__less__than_1);
    GD_GMETHOD_CASE(estring_3);
    GD_GMETHOD_CASE_DEFAULT;
  }
}

G_DEF_PRINT()
{
  G_STD_DECL;
  unsigned long size, limit, k;
  Shallow(g_self);
  size = g_self->index;
  limit = (size > g_length ? g_length : size);
  GD_PUTC('"');
  for (k = 0; k < limit; k++) {
    int c = g_self->body[k];
    if (!isgraph(c) && c!=' ') {
      GD_PUTC('\\');
      GD_PUTC('0'+((c>>6)&3));
      GD_PUTC('0'+((c>>3)&7));
      GD_PUTC('0'+(c&7));
    } else {
      if (c=='\\' || c=='"') GD_PUTC('\\');
      GD_PUTC(c);
    }
  }
  if( limit != size ) G_PRINT("..");
  GD_PUTC('"');
  G_RETURN_FROM_PRINT();
}

GDDEF_COMPARE()
{
  G_STD_DECL;
  unsigned long size1, size2, limit, k;
  Shallow(g_self);
  size1 = g_self->index;
  Shallow(GD_OTHER);
  size2 = GD_OTHER->index;
  limit = (size1 <= size2 ? size1 : size2);
  for (k = 0; k < limit; k++) {
    unsigned char elem;
    Shallow(g_self);
    elem = g_self->body[k];
    Shallow(GD_OTHER);
    if (elem != GD_OTHER->body[k]) {
      return makeint(elem > GD_OTHER->body[k] ? k+1 : -(k+1));
    }
  }
  if (size1 != size2) {
    return makeint(size1 >= size2 ? size1+1 : -(size2+1));
  } else {
    return makeint(0);
  }
}

GDDEF_HASH()
{
  G_STD_DECL;
  unsigned long size;
  Shallow(g_self);
  size = g_self->index;
  if (size == 0) {
    return makeint(0);
  } else {
    return
      makeint( 0x813 * g_self->body[0] +
		0x425 * g_self->body[size >> 1] +
		0x3c9 * g_self->body[size - 1] );
  }
}

#define GDUSE_MY_GUNIFY
#define GDUSE_MY_UNIFY
#define GDUSE_MY_PRINT
#define GDUSE_MY_GC
#ifdef DIST
#define GDUSE_MY_ENCODE
#endif
#define GDUSE_MY_GENERIC
#define GDUSE_MY_GGENERIC
#define GDUSE_MY_COMPARE
#define GDUSE_MY_HASH

/* define the method table structure of the vector */
#include <klic/gd_methtab.h>

/*  new_string function */
G_DEF_NEW()
{
  G_STD_DECL_FOR_NEW();
  G_OBJ_TYPE* newstr;
  unsigned char *body;
  long size;
  q init;

  if( g_argc != 1 ) G_ERROR_IN_NEW("Too few or too many arguments");
  init = g_argv[0];
  G_DEREF_FOR_NEW(init);
  G_SET_NEWOBJ_FOR_NEW(newstr, G_SIZE_IN_Q(G_OBJ_TYPE));
  if( isint(init) ){
    unsigned long qsize;
    size = intval(init);
    if (size < 0) G_ERROR_IN_NEW("Negative size specified");
    qsize = ROUND_UP(size);
    GD_ALLOC_AREA_FOR_NEW(body, (unsigned char *), qsize);
    BZERO(body, qsize*sizeof(q));
  }else if( init==NILATOM || iscons(init) ){
    unsigned long qsize, k;
    for (size=0; ; size++) {
      q elem;
      if (init == NILATOM) break;
      elem = car_of(init);
      G_DEREF_FOR_NEW(elem);
      if( !isint(elem) || intval(elem) < 0 || 256 <= intval(elem) ){
	G_ERROR_IN_NEW("Invalid parameter");
      }
      init = cdr_of(init);
      G_DEREF_FOR_NEW(init);
      if( init != NILATOM && !iscons(init) )
	G_ERROR_IN_NEW("Invalid parameter");
    }
    qsize = ROUND_UP(size);
    init = g_argv[0];
    GD_ALLOC_AREA_FOR_NEW(body, (unsigned char *), qsize);
    for (k=0; k<size; k++) {
      q elem;
      G_DEREF_FOR_NEW(init);
      elem = car_of(init);
      G_DEREF_FOR_NEW(elem);
      body[k] = intval(elem);
      init = cdr_of(init);
    }
  } else {
    G_ERROR_IN_NEW("Invalid parameter");
  }
  newstr->next = STRING_SHALLOW_MARK;
  newstr->iscnst = 0;
  newstr->index = size;
  newstr->body = body;
  return makefunctor(newstr);
}

#ifdef DIST
static void
decode_byte_string(inbuf)
     combuf *inbuf;
{
  G_STD_DECL;
  G_OBJ_TYPE* newstring;
  unsigned char *body;
  unsigned long size;
  long *tmp_buf;
  int i;
  q res;

  G_HEAPALLOC_WITH_CHECK(newstring, G_SIZE_IN_Q(G_OBJ_TYPE),
			 (G_OBJ_TYPE*), res);
  if(res == GENERIC_GCREQUEST){ 
      add_this_more_space(G_SIZE_IN_Q(G_OBJ_TYPE));
      inbuf->rd_index--;
      return;
  }
  newstring->method_table = &G_method_table;
  newstring->index = (long)GET_BUFFER(inbuf);

  size = ROUND_UP(newstring->index);

  {
    q *tmpbody;
    G_HEAPALLOC_WITH_CHECK(tmpbody, size, (q*), res); 
    body = (unsigned char *) tmpbody;
  }

  if(res == GENERIC_GCREQUEST){ 
      add_this_more_space(size);
      inbuf->rd_index -= 2;
      return;
  }

  tmp_buf = (long *)body;
  for(i = 0 ; i < size ; i++){
    tmp_buf[i] = GET_BUFFER(inbuf);
  }
  
  newstring->next    = STRING_SHALLOW_MARK;
  newstring->iscnst  = 0;
  newstring->body  = body;

  push_decode_stack((q)makefunctor(newstring));
}
#endif /* DIST */

extern unsigned char*
generic_string_body(G_OBJ_TYPE* str)
{
  Shallow(str);
  return (str->body);
}

extern unsigned long
generic_string_size(G_OBJ_TYPE* str)
{
  Shallow(str);
  return (str->index);
}

extern q
convert_c_string_to_klic_string(cstr)
     char *cstr;
{
  q argv[1];
  q str;
  long len = strlen(cstr);

  argv[0] = makeint(len);
  str = byte__string_g_new(1, argv);
  if( !isref(str) ){
    BCOPY(cstr, ((struct byte_string_object *)functorp(str))->body, len);
  }
  return str;
}

extern q
convert_binary_c_string_to_klic_string(cstr, len)
     char *cstr;
     long len;
{
  q argv[1];
  q str;

  argv[0] = makeint(len);
  str = byte__string_g_new(1, argv);
  if( !isref(str) ){
    BCOPY(cstr, ((struct byte_string_object *)functorp(str))->body, len);
  }
  return str;
}

extern char*
convert_klic_string_to_c_string(s)
     q s;
{
  struct byte_string_object *str =
    (struct byte_string_object *)functorp(s);
  char *cstr;
  Shallow(str);
  cstr = (char *)malloc_check(str->index+1);
  BCOPY((char *)str->body, cstr, str->index);
  cstr[str->index] = '\0';
  return cstr;
}

/* Interface with builtin */

extern q
size_of_string(s)
     q s;
{
  Shallow(STRING_OBJ(s));
  return makeint(STRING_OBJ(s)->index);
}

extern q
element_of_string(s, k)
     q s, k;
{
  Shallow(STRING_OBJ(s));
  if( intval(k) < 0 || (long) STRING_OBJ(s)->index <= intval(k) ){
    return (q) 0;
  } else {
    return makeint(STRING_OBJ(s)->body[intval(k)]);
  }
}

#undef CLASS_NAME
#define CLASS_NAME  string

G_DEF_NEW()
{
  G_STD_DECL_FOR_NEW();
  long elemsize;

  if( g_argc != 2 ) G_ERROR_IN_NEW("Too few or too many arguments");
  GDSET_INTARG_FOR_NEW(elemsize, g_argv[1]);
  if (elemsize != 8) G_ERROR_IN_NEW("Only byte strings are supported now");
  return byte__string_g_new(1, g_argv);
}
