/* ---------------------------------------------------------- 
%   (C)1995 Institute for New Generation Computer Technology 
%       (Read COPYRIGHT for detailed information.) 
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */
/* random number generator */

#include <klic/basic.h>  /* NRAND48, fatalf */

#ifdef NRAND48
#include <stdio.h>

#include <klic/struct.h>
#include <klic/g_methtab.h>
#include <klic/g_basic.h>
#include <klic/gcobject.h>

#define CLASS_NAME  random__numbers
#define G_OBJ_TYPE  struct random_number_generator
#define G_OBJ_SIZE  G_SIZE_IN_Q(G_OBJ_TYPE)

#include <klic/gc_macro.h>

G_OBJ_TYPE {
  struct consumer_object_method_table* method_table;
  long range;
  long max;
  unsigned short state[3];
};

GCDEF_UNIFY()
{
  G_STD_DECL;

 top:
  if( g_term == NILATOM ){
    GC_TERMINATE;
  }else if( isref(g_term) ){
    q temp = derefone(g_term);
    if( isref(temp) && (temp == g_term || derefone(temp) == g_term) ){
      GC_RETURN_WITH_HOOK(g_term);
    }else{
      g_term = temp;
      goto top;
    }
  }else if( iscons(g_term) ){
    long one_random;

    do{
      one_random = nrand48(g_self->state);
    }while( one_random >= g_self->max );

    derefone(car_of(g_term)) = makeint(one_random % g_self->range);
    g_term = cdr_of(g_term);
    goto top;
  }else{
    fatalf("random_numbers: does not support message %k", g_term);
  }
}

G_DEF_PRINT()
{
  G_STD_DECL;
  fprintf(g_fp, "RANDOM@%X", (q*) g_self - new_space_top());
  return 0;
}

G_DEF_GC()
{
  G_STD_DECL;
  q* newself = klic_alloc(sizeof(G_OBJ_TYPE) / sizeof(q));

  BCOPY(g_self, newself, sizeof(G_OBJ_TYPE));
  return newself;
}

#define GCUSE_MY_UNIFY
#define GCUSE_MY_PRINT
#define GCUSE_MY_GC

#include <klic/gc_methtab.h>

G_DEF_NEW()
{
  G_STD_DECL_FOR_NEW();
  G_OBJ_TYPE* obj;
  long seed;
  long range;

  if( g_argc < 1 || 2 < g_argc ){
    debug_printf("Wrong number of arguments (%d)", g_argc);
    G_ERROR_IN_NEW("");
  }

  G_DEREF_FOR_NEW(g_argv[0]);
  if( !isint(g_argv[0]) ) G_ERROR_IN_NEW("First argument is not integer");
  range = intval(g_argv[0]);
  if( range < 1 ){
    debug_printf("Invalid first argument (%d)", range);
    G_ERROR_IN_NEW("");
  }

  if( g_argc > 1 ){
    G_DEREF_FOR_NEW(g_argv[1]);
    if( !isint(g_argv[1]) ) G_ERROR_IN_NEW("Second argument is not integer");
    seed = intval(g_argv[1]);
  }else{
    seed = 0;
  }

  G_SET_NEWOBJ_FOR_NEW(obj, G_OBJ_SIZE);

  obj->state[0] = seed >> (sizeof(seed)*4);
  obj->state[1] = seed >> (sizeof(seed)*2);
  obj->state[2] = seed >> (sizeof(seed)*0);
  obj->range = range;
  obj->max = ((~0LU << 1) >> 1) / range * range;
  return GC_MAKE_HOOK_VAR((struct consumer_object*) obj);
}
#endif  /* NRAND48 */
